// z80snap.cpp
// Revision 11-oct-2006
// (C) 2003-2006 Julian Albo

#include "z80snap.h"

#include <stdexcept>

#include <math.h>
// Don't use cmath for compatibility with old compilers.

// For debugging
#include <iostream>
using std::cerr;
using std::endl;

using std::runtime_error;

using z80snap::z80byte;


namespace
{

// Length of compressed blocks.
const size_t compblock= 16 * 1024;

struct BlockHead
{
	std::streamsize length;
	int page;
	bool read (std::istream & is);
	void write (std::ostream & out);
};

bool BlockHead::read (std::istream & is)
{
	z80byte rawbh [3];
	is.read (reinterpret_cast <char *> (rawbh), 3);
	if (is.eof () )
		return false;
	if (! is || is.gcount () != 3)
		throw "Error reading block header";
	length= rawbh [0] | rawbh [1] << 8;
	page= rawbh [2];
	return true;
}

void BlockHead::write (std::ostream & out)
{
	z80byte rawbh [3];
	rawbh [0]= length % 256;
	rawbh [1]= (length / 256) % 256;
	rawbh [2]= page;
	out.write (reinterpret_cast <char *> (rawbh), 3);
}

void check (size_t n)
{
	if (n > 0xFFFF)
		throw "PEEK out of address space";
}

void showpos (std::istream & is)
{
	is.clear ();
	cerr << "At position " << is.tellg () << endl;
}

} // namespace


z80snap::Header::Header (std::istream & is, bool debug)
{
	//const std::streamsize headsize= 30;
	//z80byte rawhead [headsize];

	is.read (reinterpret_cast <char *> (rawhead), headsize);
	if (! is || std::streamsize (is.gcount () ) != headsize)
		throw ErrorReadingHeader ();

	regs.A= rawhead [0];
	regs.F= rawhead [1];
	regs.BC= rawhead [2] | rawhead [3] << 8;
	regs.HL= rawhead [4] | rawhead [5] << 8;
	regs.PC= rawhead [6] | rawhead [7] << 8;
	regs.SP= rawhead [8] | rawhead [9] << 8;

	z80byte value12= rawhead [12];
	if (value12 == 255)
		value12= 1;
	compressed= value12 & 0xA0; // Bit 5

	regs.DE= rawhead [13] | rawhead [14] << 8;
	regs.BCp= rawhead [15] | rawhead [16] << 8;
	regs.DEp= rawhead [17] | rawhead [18] << 8;
	regs.HLp= rawhead [19] | rawhead [20] << 8;
	regs.Ap= rawhead [21];
	regs.Fp= rawhead [22];
	regs.IY= rawhead [23] | rawhead [24] << 8;
	regs.IX= rawhead [25] | rawhead [26] << 8;

	oldstyle= regs.PC != 0;
	if (! oldstyle)
	{
		// Extended format

		// Get extension size.

		z80byte rawlength [2];
		is.read (reinterpret_cast <char *> (rawlength), 2);
		if (! is || is.gcount () != 2)
			throw ErrorReadingHeader ();
		//std::streamsize extlength=
		extlength=
			rawlength [0] | rawlength [1] << 8;
		if (debug)
			cerr << "Extension lenght: " << extlength << endl;
		if (extlength == 0)
			throw ErrorReadingHeader ();

		// Load header extension.

		//z80byte * rawext= new z80byte [extlength];
		rawext= new z80byte [extlength];
		is.read (reinterpret_cast <char *> (rawext), extlength);
		if (! is || static_cast <size_t> (is.gcount () ) != extlength)
		{
			delete [] rawext;
			throw ErrorReadingHeader ();
		}
		const std::streamsize extoff= headsize + 2;
		regs.PC= rawext [32 - extoff] | rawext [33 - extoff];

		if (debug)
			cerr << "Byte type: " <<
				int (rawext [34 - extoff] ) << endl;
		switch (rawext [34 - extoff] )
		{
		case 0:
			cerr << "48K\n";
			type= Type48;
			break;
		case 1:
			cerr << "48K + IF 1\n";
			type= Type48;
			break;
		case 2: // Added type 2 according to libspectrum.
			// Spectrum 48K
			cerr << "SamRam\n";
			type= Type48;
			break;
		case 3:
			cerr << "128 K\n";
			if (extlength == 23)
				type= Type128;
			else
				type= Type48;
			break;
		case 4:
			cerr << "128K\n";
			type= Type128;
			break;
		case 5:
			cerr << "128K + IF 1\n";
			type= Type128;
			break;
		case 6:
			type= Type128;
			break;
		case 7:
		case 8: // Added type 8 according to libspectrum
			// Spectrum +3
			type= Type128;
			break;
		case 12:
			// Spectrum +2
			type= Type128;
			break;
		case 13:
			// Spectrum +2A
			type= Type128;
			break;
		default:
			type= TypeUnknown;
		}

		// Spectaculator extension: if bit 7 of byte 37
		// is on the values of hardware are changed.
		if (rawext [37 - extoff] & 0x80)
		{
			if (type == Type48)
				type= Type16;
		}

		//delete [] rawext;

	}
	else
	{
		// Old format

		type= Type48;
	}

}

void z80snap::Header::save (std::ostream & out)
{
	out.write (reinterpret_cast <char *> (rawhead), headsize);
	if (! oldstyle)
	{
		z80byte rawlength [2];
		rawlength [0]= extlength % 256;
		rawlength [1]= extlength / 256;
		out.write (reinterpret_cast <char *> (rawlength), 2);
		out.write (reinterpret_cast <char *> (rawext), extlength);
	}
}

z80snap::Snapshot::Snapshot (std::istream & is, bool debug) :
	head (is, debug)
{
	z80byte discard [compblock];
	if (head.isoldstyle () )
	{
		if (head.iscompressed () )
			loadblockv1 (is, debug);
		else
		{
			if (debug)
				cerr << "Loading V1 uncompressed block" <<
					endl;
			is.read (reinterpret_cast <char *> (memory),
				48 * 1024);
			size_t r= is.gcount ();
			if (r != compblock && r != 48 * 1024)
				std::cerr << "WARNING: uncompressed block "
					" of size " << r << endl;
		}
	}
	else
	{
		// Read data blocks.

		MachineType type= head.gettype ();
		BlockHead block;
		bool none= true;
		while (block.read (is) )
		{
			none= false;
			if (debug)
			{
				cerr << "Block length: " <<
					block.length << endl <<
					"Page: " << block.page << endl;
			}
			z80byte * aux;
			switch (block.page)
			{
			case 3:
				if (type == Type48)
					aux= discard;
				else
					aux= memory + 0xC000; // Page 0
				break;
			case 4:
				if (type == Type48)
					aux= memory + 0x8000;
				else
					aux= extramemory; // Page 1
				break;
			case 5:
				if (type == Type48)
					aux= memory + 0xC000;
				else
					aux= memory + 0x8000; // Page 2
				break;
			case 6:
				if (type == Type48)
					aux= discard;
				else
					aux= extramemory + 0x4000; // Page 3
				break;
			case 7:
				if (type == Type48)
					aux= discard;
				else
					aux= extramemory + 0x8000; // Page 4
				break;
			case 8:
				if (type == Type48)
					aux= memory + 0x4000;
				else
					aux= memory + 0x4000; // Page 5
				break;
			case 9:
				if (type == Type48)
					aux= discard;
				else
					aux= extramemory + 0xC000; // Page 6
				break;
			case 10:
				if (type == Type48)
					aux= discard;
				else
					aux= extramemory + 0x10000; // Page 7
				break;
			default:
				aux= discard;
			}
			loadblock (is, aux, block.length, debug);
		}
		if (none)
			throw "There is no blocks!";
	}
}

void z80snap::Snapshot::save (std::ostream & out)
{
	head.save (out);

	// Just a little of additional space just in case...
	z80byte compr [2 * compblock];

	BlockHead block;

	size_t len;

	switch (head.gettype () )
	{
	case Type16:
		len= compressblock (memory + 0x4000, compr);
		block.length= len;
		block.page= 8;
		block.write (out);
		out.write (reinterpret_cast <char *> (compr), len);

		break;		
	case Type48:
		len= compressblock (memory + 0x8000, compr);
		block.length= len;
		block.page= 4;
		block.write (out);
		out.write (reinterpret_cast <char *> (compr), len);

		len= compressblock (memory + 0xC000, compr);
		block.length= len;
		block.page= 5;
		block.write (out);
		out.write (reinterpret_cast <char *> (compr), len);

		len= compressblock (memory + 0x4000, compr);
		block.length= len;
		block.page= 8;
		block.write (out);
		out.write (reinterpret_cast <char *> (compr), len);

		break;
	case Type128:		
		len= compressblock (memory + 0xC000, compr);
		block.length= len;
		block.page= 3;
		block.write (out);
		out.write (reinterpret_cast <char *> (compr), len);

		len= compressblock (extramemory, compr);
		block.length= len;
		block.page= 4;
		block.write (out);
		out.write (reinterpret_cast <char *> (compr), len);

		len= compressblock (memory + 0x8000, compr);
		block.length= len;
		block.page= 5;
		block.write (out);
		out.write (reinterpret_cast <char *> (compr), len);

		len= compressblock (extramemory + 0x4000, compr);
		block.length= len;
		block.page= 6;
		block.write (out);
		out.write (reinterpret_cast <char *> (compr), len);

		len= compressblock (extramemory + 0x8000, compr);
		block.length= len;
		block.page= 7;
		block.write (out);
		out.write (reinterpret_cast <char *> (compr), len);

		len= compressblock (memory + 0x4000, compr);
		block.length= len;
		block.page= 8;
		block.write (out);
		out.write (reinterpret_cast <char *> (compr), len);

		len= compressblock (extramemory + 0xC000, compr);
		block.length= len;
		block.page= 9;
		block.write (out);
		out.write (reinterpret_cast <char *> (compr), len);

		len= compressblock (extramemory + 0x10000, compr);
		block.length= len;
		block.page= 10;
		block.write (out);
		out.write (reinterpret_cast <char *> (compr), len);
		break;
	case TypeUnknown:
		throw runtime_error ("Can't save unknown type");
	}
}

void z80snap::Snapshot::poke (size_t n, z80byte value)
{
	check (n);
	memory [n]= value;
}

z80snap::z80byte z80snap::Snapshot::peek (size_t n)
{
	check (n);
	return memory [n];
}

size_t z80snap::Snapshot::peek16 (size_t n)
{
	check (n);
	return memory [n] | memory [n + 1] << 8;
}

size_t z80snap::Snapshot::peeklinenumber (size_t n)
{
	check (n);
	return memory [n]  << 8 | memory [n + 1];
}

double z80snap::Snapshot::peeknumber (size_t n)
{
	// Returns a number from the Spectrum Basic numeric format.
	z80byte b0= memory [n];
	z80byte b1= memory [n + 1];
	z80byte b2= memory [n + 2];
	z80byte b3= memory [n + 3];
	z80byte b4= memory [n + 4];

	// Special case for integers between -65535 and 65535.
	// The case when b0 and b4 are 0, b1 is 255 and b2 and b3 are 0
	// seem special. When POKEing it in the Spectrum it reports the
	// value -1E-38 but interpreted here as floating gives
	// -2.92726E-39. As last resource, I return -1E-38 directly to
	// be consistent with the Spectrum. I think that combination
	// can't be the result of a normal calculation.

	if (b0 == 0 && b4 == 0 && (b1 == 0 || b1 == 255) )
	{
		size_t n= b2 | b3 << 8;
		if (b1)
		{
			if (n)
				return -static_cast <double> (65536 - n);
			else return -1e-38;
		}
		return static_cast <double> (n);
	}

	// First bit of mantissa is always 1.
	size_t mantissa= ( (b1 & 0x7F) | 0x80) << 24 |
		b2 << 16 | b3 << 8 | b4;
	// First bit of b1 is the sign.
	bool negative= b1 & 0x80;
	// Exponent - 128, and 32 bits after the point.
	int exponent= b0 - 160;
	double result= 	pow (2, exponent) * mantissa;
	if (negative)
		result= -result;
	return result;
}

void z80snap::Snapshot::loadblock (std::istream & is,
	z80byte * ptr, std::streamsize length, bool debug)
{
	if (length == 0xFFFF)
	{
		is.read (reinterpret_cast <char *> (ptr), compblock);
		if (! is || static_cast <size_t> (is.gcount () ) != compblock)
			throw "Error reading block header";
	}
	else
	{
		size_t pos= 0;
		const z80byte mark= '\xED';
		z80byte c;
		while (length > 0)
		{
			if (pos >= compblock)
				throw "Block too long";
			c= is.get ();
			if (! is)
				throw "Error reading block";
			--length;
			if (c != mark)
			{
				* (ptr++)= c;
				++pos;
			}
			else
			{
				c= is.get ();
				if (! is)
				{
					* (ptr++)= mark;
					++pos;
					continue;
				}
				--length;
				if (c != mark)
				{
					* (ptr++)= mark;
					++pos;
					if (pos >= compblock)
						throw "Block too long";
					* (ptr++)= c;
					++pos;
					continue;
				}
				z80byte times= is.get ();
				if (! is)
					throw "Error reading block";
				--length;
				c= is.get ();
				if (! is)
					throw "Error reading block";
				--length;
				if (debug)
					cerr << "Byte " << int (c) <<
						" repeated " <<
						int (times) << endl;
				for ( ; times > 0; --times)
				{
					if (pos >= compblock)
						throw "Block too long";
					* (ptr++)= c;
					++pos;
				}
			}
		}
		if (pos != compblock)
			cerr << "WARNING: uncompressed block is not 16K long"
				<< endl;
	}
}

size_t z80snap::Snapshot::compressblock (z80byte * ptr, z80byte * compr)
{
	const z80byte mark= '\xED';
	size_t newpos= 0;
	for (size_t pos= 0; pos < compblock; )
	{
		z80byte c= ptr [pos];
		size_t minrep= c == mark ? 2 : 5;
		size_t i= 0;
		while (i < 255 && pos + i + 1 < compblock
				&& ptr [pos + i + 1] == c)
		{
			++i;
		}
		if (i >= minrep)
		{
			compr [newpos++]= mark;
			compr [newpos++]= mark;
			compr [newpos++]= static_cast <unsigned char> (i);
			compr [newpos++]= c;
			pos+= i;
		}
		else
		{
			compr [newpos++]= c;
			++pos;
			// The byte following a single ED
			// is never compressed.
			if (c == mark && pos < compblock)
			{
				compr [newpos++]= ptr [pos++];
			}
		}
	}
	// End block mark
	return newpos;
}

void z80snap::Snapshot::saveblock (std::ostream & out, z80byte * ptr)
{
	const z80byte mark= '\xED';
	for (size_t pos= 0; pos < 16 * 1024; ++pos)
	{
		z80byte c= ptr [pos];
		if (c == mark && pos < 16 * 1024 -1 && ptr [pos + 1] == mark)
		{
			out << mark << mark << '\x02' << mark;
			++pos;
		}
		else
			out << c;
	}
	// End block mark
	out << '\x00' << mark << mark << '\x00';
}

void z80snap::Snapshot::loadblockv1 (std::istream & is, bool debug)
{
	z80byte * ptr= memory + 0x4000;
	size_t pos= 0;
	size_t limit= 0xC000U + 1 + 100;
	// Can be at most 48K, 1 byte extra for end mark detection.
	const z80byte mark= '\xED';
	z80byte c;
	if (debug)
		cerr << "Loading V1 block" << endl;
	for (;;)
	{
		c= is.get ();
		if (! is)
		{
			showpos (is);
			throw "Error reading block";
		}
		if (c != mark)
		{
			if (pos >= limit)
				throw "Block too long";
			* (ptr++)= c;
			++pos;
		}
		else
		{
			c= is.get ();
			if (! is)
			{
				if (pos >= limit)
					throw "Block too long";
				* (ptr++)= mark;
				++pos;
				continue;
			}
			if (c != mark)
			{
				if (pos >= limit)
					throw "Block too long";
				* (ptr++)= mark;
				++pos;
				if (pos >= limit)
					throw "Block too long";
				* (ptr++)= c;
				++pos;
				continue;
			}
			z80byte times= is.get ();
			if (! is)
			{
				showpos (is);
				throw "Error reading block";
			}
			if (times == 0)
			{
				// End mark.
				if (pos > 1 && memory [0x4000 + pos - 2] != 0)
					cerr << "WARNING: non conforming "
						"end of block mark" << endl;
				--pos;
				break;
			}
			c= is.get ();
			if (! is)
			{
				showpos (is);
				throw "Error reading block";
			}
			if (debug)
				cerr << "Byte " << int (c) <<
					" repeated " << int (times) << endl;
			for ( ; times > 0; --times)
			{
				if (pos >= limit)
					throw "Block too long";
				* (ptr++)= c;
				++pos;
			}
		}
	}
	if (pos != 16 * 1024 && pos != 48 * 1024)
		cerr << "WARNING: uncompressed block is not 16K or 48K long,"
			" is " << pos << " bytes long" << endl;
}

// End of z80snap.cpp
